/**
 * CheapWAV.java: quick analysis of WAVE files.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-18 donfede@casagrau.org
 * credit to: Ringdroid, with a primary developer of Dominic Mazzoni, a Google employee and also author of Audacity. --  http://code.google.com/p/ringdroid/
 * $Id: CheapWAV.java,v 1.19 2013-03-31 23:02:24 donfede Exp $
 */

/*
 * ORIGINAL Ringdroid Apache/BSD Copyright by Google:
 * Copyright (C) 2008 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.ringdroid.soundfile;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Date;

import org.casagrau.mpq.mpq_android.R;

import android.content.SharedPreferences;
import android.os.Environment;
import android.preference.PreferenceManager;
import android.util.Log;

/**
 * CheapWAV represents a standard 16-bit WAV file, splitting it into
 * artificial frames of 20 ms and taking the maximum of each frame to
 * get an approximation of the waveform contour.
 */
public class CheapWAV extends CheapSoundFile {
	private static final String LOGTAG = "CheapWAV";
	
	public static Factory getFactory() {
        return new Factory() {
            public CheapSoundFile create() {
                return new CheapWAV();
            }
            public String[] getSupportedExtensions() {
                return new String[] { "wav" };
            }
        };
    }

    // Member variables containing frame info
    private int mNumFrames;
    private int mNumNoiseFrames;
    private int[] mFrameOffsets = new int [0];
    private int[] mFrameLens = new int [0];
    private int[] mFrameGains = new int [0];
    private int[] mFrameNoise = new int[0];
    private int mFrameBytes;
    private int mFileSize;
    private int mSampleRate;
    private int mBitRate;
    private int mChannels;

    public CheapWAV() {
    }

    public int getNumFrames() {
        return mNumFrames;
    }
    
    public int getNumNoiseFrames() {
    	return mNumNoiseFrames;
    }

    public int getSamplesPerFrame() {
        return mSampleRate / 50;
    }

    public int[] getFrameOffsets() {
        return mFrameOffsets;
    }

    public int[] getFrameLens() {
        return mFrameLens;
    }

    public int[] getFrameGains() {
        return mFrameGains;
    }
    
    public int[] getFrameNoise() {
    	return mFrameNoise;
    }
    

    public int getFrameBytes() {
    	return mFrameBytes;
    }

    public int getFileSizeBytes() {
        return mFileSize;        
    }

    public int getSampleRate() {
        return mSampleRate;
    }

    public int getChannels() {
        return mChannels;
    }
    
    public int getBitRate() {
    	return mBitRate;
    }

    public String getFiletype() {
        return "WAV";
    }

    public void ReadFile(File inputFile)
            throws java.io.FileNotFoundException,
                   java.io.IOException {
        super.ReadFile(inputFile); 
    	Log.d(LOGTAG, "ReadFile() 00  inputFile: " + inputFile);

        // Try to read meta data from a cached file.
        boolean lrc = ReadFileMeta(inputFile);
        if (lrc) return;
        
    	Log.d(LOGTAG, "ReadFile() 99");
    } // END ReadFile()


    /**
     * Read the state of key data members from disk.
     * @param inputParentFile File of the parent audio file.
     * @return boolean flag, true if data was read from disk.
     * This method works in conjunction with {@link #WriteFileMeta()}.
     */
    @SuppressWarnings({ "deprecation", "unused" })
	public boolean ReadFileMeta(File inputParentFile) {
    	Log.d(LOGTAG, "ReadFileMeta() 00  P File: " + inputParentFile.toString());

    	boolean lrc = false;	// Return code.

    	String lparentFileName = inputParentFile.getName();
    	String[] lcomponents = lparentFileName.split("\\.");
        if (lcomponents.length < 2) {
        	Log.e(LOGTAG, "ReadFileMeta() 15 -> 99  EXIT ignoring error, insufficient components: " + lcomponents.toString());
            return lrc;
        }
        String lmetaFileName = inputParentFile.getParent() + File.separator + "meta_dir" + File.separator + lcomponents[lcomponents.length-2] + ".meta";
    	File lMetaFile = new File(lmetaFileName);
    	long lMetaLen = lMetaFile.length();
        Log.d(LOGTAG, "ReadFileMeta() 22   M File: " + lMetaFile + " len: " + lMetaLen);

        // Check that meta file is valid.
        if ( (!lMetaFile.isFile()) || (0 == lMetaLen) ) {
        	Log.w(LOGTAG, "ReadFileMeta() 25 -> 99  EXIT ignoring error, invalid meta file");
        	return lrc;
        }
        
        //Log.d(LOGTAG, "ReadFileMeta() 30  original object: " + this.toString() );
        String lheader = "";
        String ldatestamp = "";
        String lreadFileName = "";
        int lmetaVersion = 0;
        // Credit to IBM page on java serializing --  http://www.ibm.com/developerworks/java/library/j-5things1/index.html
        try {
	        FileInputStream fis = new FileInputStream(lMetaFile);
	        ObjectInputStream ois = new ObjectInputStream(fis);
	        lheader = ois.readLine();
	        ldatestamp = ois.readLine();
	        lmetaVersion = ois.readInt();
	        
	        if (2 == lmetaVersion) {
		        lreadFileName = (String) ois.readObject();
		        mFileSize = ois.readInt();
		        mChannels = ois.readInt();
		        mSampleRate = ois.readInt();
		        mBitRate = ois.readInt();
		        mNumFrames = ois.readInt();
		        mNumNoiseFrames = ois.readInt();
                if (mProgressListener != null) {	// FIXME: this listener check sure feels like a function...
                    boolean keepGoing = mProgressListener.reportProgress(0.10);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameOffsets = (int[]) ois.readObject();
		        if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.25);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameLens = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.50);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameGains = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.75);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
		        mFrameNoise = (int[]) ois.readObject();
                if (mProgressListener != null) {
                    boolean keepGoing = mProgressListener.reportProgress(0.98);
                    if (!keepGoing) {
                        return lrc;
                    }
                }
	        } else {
	        	Log.w(LOGTAG, "ReadFileMeta() 37 -> 99  EXIT ignoring error, invalid meta file version: " + lmetaVersion);
	        	return lrc;
	        }
	        ois.close();
	        lrc = true;
        } catch (Exception e) {
            Log.e(LOGTAG, "ReadFileMeta() Faile to read meta file: " + lMetaFile.toString() + " exception ex: " + e.getLocalizedMessage() );
            e.printStackTrace();
        }        
//        Log.d(LOGTAG, "ReadFileMeta() 41 header: " + lheader);
//        Log.d(LOGTAG, "ReadFileMeta() 43 datestamp: " + ldatestamp);
//        Log.d(LOGTAG, "ReadFileMeta() 44 version: " + lmetaVersion);
//        Log.d(LOGTAG, "ReadFileMeta() 45 read filename: " + lreadFileName);
//        Log.d(LOGTAG, "ReadFileMeta() 47  read object: " + this.toString() );

        return lrc;
    } // END WriteFileMeta(File inputParentFile)
    
    
    
    public void WriteFile(File outputFile, int startFrame, int numFrames)
            throws java.io.IOException {
    	/*
        outputFile.createNewFile();
        FileInputStream in = null; // new FileInputStream(inputFile);
        FileOutputStream out = new FileOutputStream(outputFile);

        long totalAudioLen = 0;
        for (int i = 0; i < numFrames; i++) {
            totalAudioLen += mFrameLens[startFrame + i];
        }

        long totalDataLen = totalAudioLen + 36;
        long longSampleRate = mSampleRate;
        long byteRate = mSampleRate * 2 * mChannels;

        byte[] header = new byte[44];
        header[0] = 'R';  // RIFF/WAVE header
        header[1] = 'I';
        header[2] = 'F';
        header[3] = 'F';
        header[4] = (byte) (totalDataLen & 0xff);
        header[5] = (byte) ((totalDataLen >> 8) & 0xff);
        header[6] = (byte) ((totalDataLen >> 16) & 0xff);
        header[7] = (byte) ((totalDataLen >> 24) & 0xff);
        header[8] = 'W';
        header[9] = 'A';
        header[10] = 'V';
        header[11] = 'E';
        header[12] = 'f';  // 'fmt ' chunk
        header[13] = 'm';
        header[14] = 't';
        header[15] = ' ';
        header[16] = 16;  // 4 bytes: size of 'fmt ' chunk
        header[17] = 0;
        header[18] = 0;
        header[19] = 0;
        header[20] = 1;  // format = 1
        header[21] = 0;
        header[22] = (byte) mChannels;
        header[23] = 0;
        header[24] = (byte) (longSampleRate & 0xff);
        header[25] = (byte) ((longSampleRate >> 8) & 0xff);
        header[26] = (byte) ((longSampleRate >> 16) & 0xff);
        header[27] = (byte) ((longSampleRate >> 24) & 0xff);
        header[28] = (byte) (byteRate & 0xff);
        header[29] = (byte) ((byteRate >> 8) & 0xff);
        header[30] = (byte) ((byteRate >> 16) & 0xff);
        header[31] = (byte) ((byteRate >> 24) & 0xff);
        header[32] = (byte) (2 * mChannels);  // block align
        header[33] = 0;
        header[34] = 16;  // bits per sample
        header[35] = 0;
        header[36] = 'd';
        header[37] = 'a';
        header[38] = 't';
        header[39] = 'a';
        header[40] = (byte) (totalAudioLen & 0xff);
        header[41] = (byte) ((totalAudioLen >> 8) & 0xff);
        header[42] = (byte) ((totalAudioLen >> 16) & 0xff);
        header[43] = (byte) ((totalAudioLen >> 24) & 0xff);
        out.write(header, 0, 44);

        byte[] buffer = new byte[mFrameBytes];
        int pos = 0;
        for (int i = 0; i < numFrames; i++) {
            int skip = mFrameOffsets[startFrame + i] - pos;
            int len = mFrameLens[startFrame + i];
            if (skip < 0) {
                continue;
            }
            if (skip > 0) {
                in.skip(skip);
                pos += skip;
            }
            in.read(buffer, 0, len);
            out.write(buffer, 0, len);
            pos += len;
        }

        in.close();
        out.close();
    */
    } // END WriteFile()
    
    
    /**
     * Return a string representation of this object.
     * @return String with description of object contents.
     */
    public String toString() {
        String lstr = new String();    
    	lstr = lstr + " mFileSize: " + mFileSize;
    	lstr = lstr + " mChannels: " + mChannels;
    	lstr = lstr + " mSampleRate: " + mSampleRate;
    	lstr = lstr + " mBitRate: " + mBitRate;
    	lstr = lstr + " mNumFrames: " + mNumFrames;
    	lstr = lstr + " nNumNoiseFrames: " + mNumNoiseFrames;
    	lstr = lstr + " mFrameOffsets[" + mFrameOffsets.length + "]: " + mFrameOffsets.toString();
    	lstr = lstr + " mFrameLens[" + mFrameLens.length + "]: " + mFrameLens.toString();
    	lstr = lstr + " mFrameGains[" + mFrameGains.length + "]: " + mFrameGains.toString();
    	lstr = lstr + " mFrameNoise[" + mFrameNoise.length + "]: " + mFrameNoise.toString();
    	return lstr;
    } // END toString()
    
} // END class CheapWAV

