/**
 * MPQAAThread.java: MPQ Audio Analyzer.
 *            Copyright (C) 2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2013-01-27 donfede@casagrau.org
 * $Id: MPQAAThread.java,v 1.7 2013-03-31 23:02:24 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.util.Date;

import org.casagrau.mpq.mpq_android.R;

import android.content.Context;
import android.content.SharedPreferences;
import android.os.Environment;
import android.os.Handler;
import android.preference.PreferenceManager;
import android.util.Log;

public class MPQAAThread extends MPQAudioWorkerThread {
	private static final String LOGTAG = "MPQAAThread";
	
    // Ported Member variables from CheapWAV containing frame info.
    private int mNumFrames;
    private int mNumNoiseFrames;
    private int[] mFrameOffsets = new int [0];
    private int[] mFrameLens = new int [0];
    private int[] mFrameGains = new int [0];
    private int[] mFrameNoise = new int[0];
    private int mFrameBytes;
    private int mFileSize;
    private int mSampleRate;
    private int mBitRate;
    private int mChannels;

	
	/** 
	 * Constructor for the worker thread.
	 * {@link #MPQAudioWorkerThread}  
	 */
	public MPQAAThread(Handler parentHandler, String filename, long filepos, int state, Context context) {
		super(parentHandler, filename, filepos, state, context);
	}
	

	/**
	 * Perform 'basic' audio analysis.
	 * 
	 */
	public void run() {
		Log.d(LOGTAG, "runAA() 00 filename: " + mFilename + " len: " + mFilelen + " noise " + mFilenoise);

		// Extract statistics from a CheapSoundFile object.
    	long lfileMS = 0;
    	long lnoiseMS = 0;
		try {
			String lFileName = Environment.getExternalStorageDirectory().getAbsolutePath() + mFilename;
			
	        File linputFile = new File(lFileName);
	        mFileSize = (int)linputFile.length();
	        if (mFileSize < 128) {
	            throw new java.io.IOException("File too small to parse");
	        }

	        FileInputStream stream = new FileInputStream(linputFile);
	        int lOffset = 0;
	        byte[] header = new byte[12];
	        stream.read(header, 0, 12);
	        lOffset += 12;
	        if (header[0] != 'R' ||
	            header[1] != 'I' ||
	            header[2] != 'F' ||
	            header[3] != 'F' ||
	            header[8] != 'W' ||
	            header[9] != 'A' ||
	            header[10] != 'V' ||
	            header[11] != 'E') {
	            throw new java.io.IOException("Not a WAV file");
	        }

	        mChannels = 0;
	        mSampleRate = 0;
	        mBitRate = 0;
	        // Chunk read loop.
	        while (lOffset + 8 <= mFileSize) {
	            byte[] chunkHeader = new byte[8];
	            stream.read(chunkHeader, 0, 8);
	            lOffset += 8;

	            int chunkLen =
	                ((0xff & chunkHeader[7]) << 24) |
	                ((0xff & chunkHeader[6]) << 16) |
	                ((0xff & chunkHeader[5]) << 8) |
	                ((0xff & chunkHeader[4]));
	            
	            // Read fmt chunk.
	            if (chunkHeader[0] == 'f' &&
	                chunkHeader[1] == 'm' &&
	                chunkHeader[2] == 't' &&
	                chunkHeader[3] == ' ') {
	                if (chunkLen < 16 || chunkLen > 1024) {
	                    throw new java.io.IOException(
	                        "WAV file has bad fmt chunk");
	                }

	                byte[] fmt = new byte[chunkLen];
	                stream.read(fmt, 0, chunkLen);
	                lOffset += chunkLen;

	                int format =
	                    ((0xff & fmt[1]) << 8) |
	                    ((0xff & fmt[0]));
	                mChannels =
	                    ((0xff & fmt[3]) << 8) |
	                    ((0xff & fmt[2]));
	                mSampleRate =
	                    ((0xff & fmt[7]) << 24) |
	                    ((0xff & fmt[6]) << 16) |
	                    ((0xff & fmt[5]) << 8) |
	                    ((0xff & fmt[4]));
	                // 4n AvgBytesPerSec [11 -8]
	                // 2n BlockAlign [13-12]
	                mBitRate = 
	                	((0xff & fmt[15]) << 8) |
	                	((0xff & fmt[14]));

	                if (format != 1) {
	                    throw new java.io.IOException(
	                        "Unsupported WAV file encoding");
	                }

	            // Read data chunk.
	            } else if (chunkHeader[0] == 'd' &&
	                       chunkHeader[1] == 'a' &&
	                       chunkHeader[2] == 't' &&
	                       chunkHeader[3] == 'a') {
	                if (mChannels == 0 || mSampleRate == 0) {
	                    throw new java.io.IOException(
	                        "Bad WAV file: data chunk before fmt chunk");
	                }

	                int frameSamples = (mSampleRate * mChannels) / 50;
	                mFrameBytes = frameSamples * (mBitRate / 8);
	            	Log.d(LOGTAG, "ReadFile() 35  sampleR: " + mSampleRate + " chan: " + mChannels + " bitR: " + mBitRate);
	            	Log.d(LOGTAG, "ReadFile() 37  framesamples: " + frameSamples + " frameBytes: " + mFrameBytes);

	                mNumFrames = (chunkLen + (mFrameBytes - 1)) / mFrameBytes;
	                mNumNoiseFrames = 0;
	                mFrameOffsets = new int[mNumFrames];
	                mFrameLens = new int[mNumFrames];
	                mFrameGains = new int[mNumFrames];
	                mFrameNoise = new int[mNumFrames];
	                
	                byte[] oneFrame = new byte[mFrameBytes];

	                int i = 0;
	                int frameIndex = 0;
	                
	                // Preparations for basic 'noise' calculations.
	    			long NOISE_LIMIT_TIME = 500;
	    			int NOISE_LIMIT_VOL = 100;
	    			// Get audio directory from preferences.
	    			if (null != mServiceContext) {
		    			SharedPreferences sharedPref = PreferenceManager.getDefaultSharedPreferences(mServiceContext);
		    			String lstrv = sharedPref.getString(mServiceContext.getString(R.string.pref_noise_vol_limit_key), mServiceContext.getString(R.string.pref_noise_vol_limit_default));
		    			String lstrt = sharedPref.getString(mServiceContext.getString(R.string.pref_noise_time_limit_key), mServiceContext.getString(R.string.pref_noise_time_limit_default));
		    			NOISE_LIMIT_VOL = Integer.parseInt(lstrv);
		    			NOISE_LIMIT_TIME = Long.parseLong(lstrt);
		    			} else Log.w(LOGTAG, "ReadFile() 39  NULL context, unable to get preferences, using defaults.");

	    			Log.d(LOGTAG, "ReadFile() 40  vol limit: " + NOISE_LIMIT_VOL + " time limit: " + NOISE_LIMIT_TIME);

	    			
	    			long lcursorTimeMS = 0;
	    			long lnoiseLastTimeMS = 0;
	                ///////////////////////

	                // Loop through and read 'artificial 20ms' frames, storing maxGain along the way.
	                while (i < chunkLen) {
	                    int oneFrameBytes = mFrameBytes;
	                    if (i + oneFrameBytes > chunkLen) {
	                        i = chunkLen - oneFrameBytes;
	                    }

	                    stream.read(oneFrame, 0, oneFrameBytes);

	                    // Loop through 'frame', byte by byte, calculating maxGain.
	                    int maxGain = 0;
	                    for (int j = 1; j < oneFrameBytes; j += 4 * mChannels) {
	                        int val = java.lang.Math.abs(oneFrame[j]);
	                        if (val > maxGain) {
	                            maxGain = val;
	                        }
	                    }

	                    mFrameOffsets[frameIndex] = lOffset;
	                    mFrameLens[frameIndex] = oneFrameBytes;
	                    mFrameGains[frameIndex] = maxGain;
	                    
	                    // Basic 'noise' calculations.  Checking for 'noise' > $VolumeLevel, within $Time/Measures (~4-8seconds).
	                    lcursorTimeMS = ((long)i * 1000) / mChannels / (mBitRate / 8) / mSampleRate;                    		
	                    if (maxGain > NOISE_LIMIT_VOL) {
	                    	// FIXME: Turning the noise flag on, should probably be throttled (log()?? or possibly retroactively) 
	                    	mFrameNoise[frameIndex] = 1;  // FIXME: 'real noies' keep this code and use constants, or toss it.
	                    	mNumNoiseFrames++;
	                    	lnoiseLastTimeMS = lcursorTimeMS;
	                    } else {
	                		mFrameNoise[frameIndex] = 0;

	                    	if (frameIndex > 1) {
		                    	// Check last frame if noise flag was on, and disable if appropriate.
		                    	if (1 <= mFrameNoise[frameIndex-1]) {
		                    		if ((lcursorTimeMS - lnoiseLastTimeMS) > NOISE_LIMIT_TIME)  {
		                    			mFrameNoise[frameIndex-1] = 0;
		                    			mNumNoiseFrames--;
			                    		// FIXME: create a for loop , checking last 8ms of frames, and disable their noise flags if less than vol limit.
		                    		} else {
			                        	mFrameNoise[frameIndex] = 2; // FIXME: 'lingering/rest noise' keep this code and use constants, or toss it.
			                        	mNumNoiseFrames++;
		                    		}
		                    	}
	                    	} // END if (frameIndex > 1)
	                    }
	                	//Log.d(LOGTAG, "ReadFile() 43  i: " + i + " lcursorTimeMS: " + lcursorTimeMS + " lastTimeMS: " + lnoiseLastTimeMS + " maxGain: " + maxGain + " noiseFrames: " + mNumNoiseFrames);

	                    frameIndex++;
	                    lOffset += oneFrameBytes;
	                    i += oneFrameBytes;

	                } // END while Loop through and read data chunk frames
	            	Log.d(LOGTAG, "ReadFile() 49  i: " + i + " lcursorTimeMS: " + lcursorTimeMS + " numFrame: " + mNumFrames + " noise: " + mNumNoiseFrames);
	            // Skip over unknown chunks.
	            } else {
	                stream.skip(chunkLen);
	                lOffset += chunkLen;
	            }
	        } // END while chunk read loop.     
	        stream.close();
	        
	        WriteFileMeta(linputFile);
						
//			int lsamplesPFrame = lcheapSound.getSamplesPerFrame(); // FIXME: consider tossing this...			
//			Log.w(LOGTAG, "runAA() 32 mNumFrames: " + mNumFrames + " mNumNoiseFrames: " + mNumNoiseFrames + " mFrameBytes: " + mFrameBytes + " sampPFrame: " + lsamplesPFrame + " mChannels: " + mChannels + " mBitRate: " + mBitRate + " mSampleRate|freq: " + mSampleRate);
			lfileMS = ((long)mNumFrames * mFrameBytes * 1000) / mChannels / (mBitRate / 8) / mSampleRate;
			lnoiseMS = ((long)mNumNoiseFrames * mFrameBytes * 1000) / mChannels / (mBitRate / 8) / mSampleRate;  		
		} catch (final Exception e) {
			Log.e(LOGTAG, "runAA() Exception: " + e.getLocalizedMessage() );
			e.printStackTrace();
			// FIXME: do we handle this exception differently than returning early (without sending AA results)??
			return;
		}
		
		// Send updates with AA results.  Report results back (from child thread to parent service; from service to SessionDetailFragment to DB).    	
    	sendPMsgAA(mFilename, lfileMS, lnoiseMS);
		Log.d(LOGTAG, "run() 99 filename: " + mFilename + " len: " + lfileMS + " noise " + lnoiseMS);
		return;
	} // END run()
	
    
    /**
     * Save the state of key data members to disk for later quick retrieval.
     * @param inputParentFile File of the parent audio file.
     * This method works in conjunction with {@link #ReadFileMeta()}.
     */
    public void WriteFileMeta(File inputParentFile) {
    	Log.d(LOGTAG, "WriteFileMeta() 00  P File: " + inputParentFile.toString());

    	String lparentFileName = inputParentFile.getName();
    	String[] lcomponents = lparentFileName.split("\\.");
        if (lcomponents.length < 2) {
        	Log.e(LOGTAG, "WriteFileMeta() 15 -> 99  EXIT ignoring error, insufficient components: " + lcomponents.toString());
            return;
        }

		File lFile = new File(inputParentFile.getParent() + File.separator + "meta_dir");
	    if (!lFile.isDirectory()) {
		    Log.e(LOGTAG, "WriteFileMeta() 16 DIR not directory - del");
	    	lFile.delete();
	    }
	    if (!lFile.exists() ) {
		    Log.w(LOGTAG, "WriteFileMeta() 17 DIR no exists - mkdir");
	    	lFile.mkdirs();
	    }
        
        String lmetaFileName = inputParentFile.getParent() + File.separator + "meta_dir" + File.separator + lcomponents[lcomponents.length-2] + ".meta";
    	File lMetaFile = new File(lmetaFileName);
        Log.d(LOGTAG, "WriteFileMeta() 22   M File: " + lMetaFile);

        // Credit to IBM page on java serializing --  http://www.ibm.com/developerworks/java/library/j-5things1/index.html
        try {
	        FileOutputStream fos = new FileOutputStream(lMetaFile);
	        ObjectOutputStream oos = new ObjectOutputStream(fos);
	        oos.writeChars("# MPQ APP meta data\n");
	        oos.writeChars("# " + new Date(System.currentTimeMillis())+ "\n");
	        oos.writeInt(2); // Current version of the MPQ meta file format.  // FIXME: this should probably go into a constant. 
	        oos.writeObject(lmetaFileName);
	        oos.writeInt(mFileSize);
	        oos.writeInt(mChannels);
	        oos.writeInt(mSampleRate);
	        oos.writeInt(mBitRate);
	        oos.writeInt(mNumFrames);
	        oos.writeInt(mNumNoiseFrames);
	        oos.writeObject(mFrameOffsets);
	        oos.writeObject(mFrameLens);
	        oos.writeObject(mFrameGains);
	        oos.writeObject(mFrameNoise);
	        oos.close();
        } catch (Exception e) {
			throw new IllegalStateException("Failed to create " + lMetaFile.toString() + " exception ex: " + e.getLocalizedMessage());
		}
        //Log.d(LOGTAG, "WriteFileMeta() 99");
    } // END WriteFileMeta(File inputParentFile)
    
    
		
} // END class MPQAAThread
