/**
 * MPQAudioFragment.java: Fragment to tie together audio related components, 
 * 							presenting UI widgets, including a waveform View, 
 * 							and starting a service to record and play audio using worker threads.
 *            Copyright (C) 2012,2013 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-12-17 donfede@casagrau.org
 * credit to: Ringdroid, with a primary developer of Dominic Mazzoni, a Google employee and also author of Audacity. --  http://code.google.com/p/ringdroid/
 *
 * $Id: MPQAudioFragment.java,v 1.40 2013-03-31 23:05:38 donfede Exp $
 */

package org.casagrau.mpq.audio_service;

import org.casagrau.mpq.mpq_android.R;
import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.app.Activity;
import android.app.ProgressDialog;
import android.content.ComponentName;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.ServiceConnection;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.os.IBinder;
import android.support.v4.app.Fragment;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.SeekBar;
import android.widget.TextView;

import com.ringdroid.soundfile.CheapSoundFile;

/**
 * Fragment to tie together audio related components.  Presents the UI widgets, 
 * including a waveform View (ported from the Ringdroid project), and uses a 
 * foreground running android service, to record and play audio.  Service worker threads 
 * do the actual audio processing, for an optimal experience that does not slow 
 * down the UI thread.
 */
public class MPQAudioFragment extends Fragment {
	private static final String LOGTAG = "MPQAudioFragment";
	
	private Activity mAct;				// Caching variable with the activity, for context and related methods.
    private Handler mFragHandler;		// Handler for the Fragment.  // FIXME: not clear if this is shared with the Activity or not... (same thread, but ...?) -- alternately, do we grab the handler from the activity???

	// Audio UI components
	private Button mPracticeButton; 	// buttonPracticeToggle
	private Button mTCButton; 			// buttonAUI_TC -- transport control: record, play, pause toggle
	private Button mStopButton;			// buttonAUI_STOP
	private SeekBar mVolControl;		// widget to set the volume (either speaker or mic, depending on the state)
	private TextView mLabelAudioPos;	// label for the cursor position
	private TextView mLabelAudioLen;	// label_total_counter , representing the total file length (in hh:mm:ss)
	private TextView mLabelVolControl; 	// label for the volume control (ex: mic or volume)
	@SuppressWarnings("unused") // Unused due to volume label currently being disabled.
	private TextView mLabelVolValue;	// label to display the current volume value
	
	// Member attributes used update UI widgets, keeping track of state and values of the audio service. 
	private MPQAudioService mAudioService;		// Handle to the audio service, received from a binding connection.
	private boolean mAudioServiceBound = false;	// Flag to track if the audio service is bound.
	private int mAudioServiceState = MPQAudioService.FLG_SRV_STATE_INIT;// Audio service state.
	private String mAudioServiceFileName;								// String with filename of the current audio managed by the audio service.
	private long mAudioServiceMSPos;									// Audio service cursor position, in milliseconds.
	private long mAudioServiceMSLen;									// Audio service *Audio* length, in milliseconds.
	private int mAudioServiceVol;						

	// Member attributes related to the audio WaveformView.
	private WaveformView mAudioWaveformView;// Handle to the waveform widget.
    private int mAudioWaveformWidth;		// Track the width of the waveform.
    private int mAudioWaveformZoom;			// Track the user selected zoom level.
    private int mAudioWaveformOffset;
    private int mAudioWaveformOffsetGoal;
    private int mAudioWaveformStartPos;
    private int mAudioWaveformEndPos;
    private int mAudioWaveformMaxPos; // FIXME: should probably be long... and document this (once it works...)
    private float mAudioWaveformDensity;
    private CheapSoundFile mAudioWaveformSoundFile;	// Handle to "CheapSoundFile", used to by waveform widget.
    private long mAudioWaveformLoadingLastUpdateTime;
    private boolean mAudioWaveformLoadingKeepGoing;
    private ProgressDialog mAudioWaveformProgressDialog;
    // More member attributes related to WaveformView; for gesture detection.
	private boolean mAudioWaveformTouchDragging;
	private float mAudioWaveformTouchStart;
	private long mAudioWaveformTouchStartMsec;
	private int mAudioWaveformTouchInitialOffset;
	private float mAudioWaveformFlingVelocity;

	
	
	public MPQAudioFragment() {
		super();
        Log.d(LOGTAG, "CONSTRUCTOR() 00");
	}
	
	@Override
	public void onAttach(Activity activity) {
		super.onAttach(activity);
		Log.d(LOGTAG, "onAttach() 00 getting activity");
		mAct = activity;
	}
	
    @Override
    public View onCreateView(LayoutInflater inflater, ViewGroup container,
                             Bundle savedInstanceState) {
    	super.onCreateView(inflater, container, savedInstanceState);
		Log.d(LOGTAG, "onCreateView() 00");
        mFragHandler = new Handler();
        
        // Inflate the layout for this fragment
        return inflater.inflate(R.layout.fragment_audio, container, false);
    }
	
    @Override
    public void onStart() {
    	super.onStart();
    	Log.d(LOGTAG, "onStart() 00");
    	Intent intent = new Intent(mAct, MPQAudioService.class);
		intent.setAction(MPQAudioService.ACTION_BIND);
		mAct.bindService(intent, mAServiceConnection, Context.BIND_AUTO_CREATE);
    	Log.d(LOGTAG, "onStart() 99");
    }
	
    @Override
    public void onResume() {
    	super.onResume();
    	Log.d(LOGTAG, "onResume() 00");		
    	
		// Bind Audio UI widget listeners.
		mPracticeButton = (Button) mAct.findViewById(R.id.buttonPracticeToggle);
		mPracticeButton.setOnClickListener(clickerPractice);
		mTCButton = (Button) mAct.findViewById(R.id.buttonAUI_TC);
		mTCButton.setOnClickListener(clickerTC);		
		mStopButton = (Button) mAct.findViewById(R.id.buttonAUI_STOP);
		mStopButton.setOnClickListener(clickerStop);
		mVolControl = (SeekBar) mAct.findViewById(R.id.seek_vol_control);
		mVolControl.setVisibility(View.GONE);
		//mVolControl.setOnSeekBarChangeListener(seekerVolControl);
		mLabelAudioPos = (TextView) mAct.findViewById(R.id.label_cursor_counter);
		mLabelAudioLen = (TextView) mAct.findViewById(R.id.label_total_counter);
		mLabelVolControl = (TextView) mAct.findViewById(R.id.label_vol_control);
		mLabelVolValue = (TextView) mAct.findViewById(R.id.label_vol_value);
		
        // Reset waveform variables, create widget, and set listener.
        mAudioWaveformSoundFile = null;
        mAudioWaveformMaxPos = 0;

        mAudioWaveformView = (WaveformView) mAct.findViewById(R.id.wave_form);
        mAudioWaveformView.setListener(waveformListener);
    	Log.d(LOGTAG, "onResume() 99");		
    } // END onResume()
      
    @Override
    public void onStop() {
    	Log.d(LOGTAG, "onStop() 00 bound_flag: " + mAudioServiceBound + " this: " + this);
    	Log.d(LOGTAG, "onStop() 05 mAudioService: " + mAudioService );		
    	super.onStop();
    	if (null != mAudioService) {
    		int serviceState = mAudioService.getStateBnd();
    		if ( (MPQAudioService.FLG_SRV_STATE_PLAY != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_REC != serviceState) &&
    				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC != serviceState) ) {
    			Intent intent = new Intent(mAct, MPQAudioService.class);
            	mAct.stopService(intent);
            }
    	}
    	if (mAudioServiceBound) {
    		mAct.unbindService(mAServiceConnection);
    	}
    } // END onStop()
    
    

	/*****************************************
	 * Private helper bind-connection callbacks and other Listeners.
	 */
	
    /**
     * Callbacks for service binding, passed to bindService().
     */
    private ServiceConnection mAServiceConnection = new ServiceConnection() {
    	public void onServiceConnected(ComponentName className, IBinder service_binder) {
    		MPQAudioService.LocalBinder binder = (MPQAudioService.LocalBinder) service_binder;
    		mAudioService = binder.getService();
    		mAudioService.setListenerBnd(aserviceListener, getTag());
    		mAudioServiceBound = true;
    		
    		mAudioServiceState = mAudioService.getStateBnd();
    		updateWidgetsLook();
			Log.d(LOGTAG, "mAServiceConnection onServiceConnected() 99 got State: " + mAudioService.getStateBnd());
    	} // END onServiceConnected()
    	public void onServiceDisconnected(ComponentName className) {
    		if (null != mAudioService) {
    			mAudioService.delListenerBnd(getTag());
    		}
    		mAudioServiceBound = false;
			Log.d(LOGTAG, "mAServiceConnection onServiceDisconnected() 99 classname: " + className);
    	}
    }; // END ServiceConnection mAServiceConnection
    
    
    
	/**
	 * MPQ Audio Service Listener.
	 */
	private MPQAudioService.MPQAudioServiceListener aserviceListener = new MPQAudioService.MPQAudioServiceListener() {
		public void onMPQAudioServiceStateChange(int newState) {
			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceStateChange() 00  new State: " + newState);
			// Trigger UI updates.	
			mAudioServiceState = newState;
			updateWidgetsLook();
		} // END onMPQAudioServiceStateChange();
		
		public void onMPQAudioServiceAudioChange(long newPos, long fileLen, int newVol) {
			Log.d(LOGTAG, "aserviceListener  onMPQAudioServiceAudioChange() 00  new Pos: " + newPos + " new Vol: " + newVol);
			// for now - Update cursor position text , and or total length.
			// later - update mic volume - possibly audio wave position or pic
			boolean lupdateFlag = false;
			if ( (newPos >= 0) && (newPos != mAudioServiceMSPos) ) {
				mAudioServiceMSPos = newPos;
				lupdateFlag = true;				
			}
			if ( (fileLen >= 0) && (fileLen != mAudioServiceMSLen) ) {
				mAudioServiceMSLen = fileLen;
				lupdateFlag = true;
			}
			if ( (newVol >= 0) && (newVol != mAudioServiceVol) ) {
				mAudioServiceVol = newVol;
				lupdateFlag = true;
			}

			if (true == lupdateFlag) 
				updateAudioWidgetsLook();
		} // END onMPQAudioServiceAudioChange()
		
		public void onMPQAudioServiceAADone(String filename, long audioLenMS, long noiseLenMS) {
			Log.w(LOGTAG, "aserviceListener  onMPQAudioServiceAADone() 00 PASS  filename: " + filename + " audio: " + audioLenMS + " noise: " + noiseLenMS);
			// Update the Waveform, by resetting the service filename.
			mAudioServiceFileName = "";
			updateWidgetsLook();
		} // END onMPQAudioServiceAADone()
	}; // END MPQAudioService.MPQAudioServiceListener aserviceListener

    

	/*****************************************
	 * Private button Listeners, for transport controls: TC/Play/Pause/Rec, Stop, Practice/Rec, Volume seek bar.
	 */
	
	/**
	 * TC session button listener.
	 * Button used for transport control: record, play, pause toggle buttons.
	 */
	private OnClickListener clickerTC = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "TC onClick() 00  state: " + mAudioServiceState);
			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
			if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) ||
					(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ) {
				intent.setAction(MPQAudioService.ACTION_PAUSE_REC);
			} else if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) {
				intent.setAction(MPQAudioService.ACTION_PLAY);
			} else if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) ||
					(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) ) {
				intent.setAction(MPQAudioService.ACTION_PAUSE_PLAY);
			}
			mAct.startService(intent);
			waveformUpdateDisplay();
			Log.d(LOGTAG, "TC onClick() 99");
		}
	}; // END OnClickListener clickerTC
	
	/**
	 * Stop session button listener.
	 */
	private OnClickListener clickerStop = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "STOP onClick() 00  state: " + mAudioServiceState);
			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
			intent.setAction(MPQAudioService.ACTION_STOP);
			mAct.startService(intent);
			mAudioServiceMSPos = 0;
			waveformUpdateDisplay();
			Log.d(LOGTAG, "STOP onClick() 99");
		}
	}; // END OnClickListener clickerStop
	
	/**
	 * Practice session button listener.
	 * If currently recording (or paused from recording) then this button will stop; 
	 * otherwise it starts a new recording.
	 */
	private OnClickListener clickerPractice = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "PRACTICE/REC onClick() 00  state: " + mAudioServiceState);
			Log.w(LOGTAG, "PRACTICE/REC onClick() 10  test delete audio **** state: " + mAudioServiceState );

// FIXME: DEBUG CODE - testing here            	
        	String lFilename = mAudioServiceFileName;
        	Log.d(LOGTAG, "PRACTICE/REC onClick() 21 to send delete audio only service intent - delete: " + lFilename);
        	
       		// Send delete Intent to the audio service.
			Intent intent = new Intent(mAct, MPQAudioService.class);
			intent.setAction(MPQAudioService.ACTION_DEL);
			// Pass filename via "extra-bundle"  
        	Bundle lextraArgs = new Bundle();
        	lextraArgs.putString(MPQAudioService.ARG_FILENAME, lFilename);
        	intent.putExtras(lextraArgs);
        	lextraArgs.clear();
        	lextraArgs.putBoolean(MPQAudioService.ARG_DEL_AUDIO_ONLY, true);
        	intent.putExtras(lextraArgs);
			mAct.startService(intent);
			
			// NOTE: Practice widget is currently disabled/removed (v0.5.3).
//			Intent intent = new Intent(v.getContext(), MPQAudioService.class);
//			// If currently recording (or paused from recording) then this button will stop; otherwise it starts a new recording.
//			if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) ||
//					(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ){
//				intent.setAction(MPQAudioService.ACTION_STOP);
//			} else {
//				intent.setAction(MPQAudioService.ACTION_REC);
//
//				// Pass filename via "extra-bundle"  
//	        	String lFilename = "/mpq/testAudioService01.wav";
//	        	Bundle lextraArgs = new Bundle();
//	        	lextraArgs.putString(MPQAudioService.ARG_FILENAME, lFilename);
//	        	intent.putExtras(lextraArgs);
//			}
//			mAct.startService(intent);
//			Log.d(LOGTAG, "PRACTICE/REC onClick() 99");
		}
	}; // END OnClickListener clickerPractice

	
	/**
	 * Volume control SeekBar listener.
	 * NOTE: *** NOT CONNECTED TO LISTENER ABOVE ***
	 */
	/** NOTE: *** NOT CONNECTED TO LISTENER ABOVE ***
	private SeekBar.OnSeekBarChangeListener seekerVolControl = new SeekBar.OnSeekBarChangeListener() {
		public void onProgressChanged(SeekBar seekBar, int progress, boolean fromUser) {
			Log.d(LOGTAG, "seekerVolC onProgressChanged() 00  progress: " + progress);
			// stop the transport... stop foreground service...
			mLabelVolValue.setText(String.valueOf(progress));
			Log.d(LOGTAG, "seekerVolC onProgressChanged() 99");
		}	
		public void onStartTrackingTouch(SeekBar seekBar) {
			Log.d(LOGTAG, "seekerVolC onStartTrackingTouch() 00  ");
			// nothing for now		
		}
		public void onStopTrackingTouch(SeekBar seekBar) {
			Log.d(LOGTAG, "seekerVolC onStopTrackingTouch() 00  ");
			// nothing for now
		}
	}; // END SeekBar.OnSeekBarChangeListener seekerVolControl
	*/
	
	
	/*****************************************
	 * Private helper methods to update UI: look of buttons, audio widgets.
	 */
	
    /**
     * Update various Audio UI widgets text and state.
     * {@link #mAudioServiceState} is checked for the current state.
     * {@link #mAudioServiceMSLen} is updated if a new length is found.
     * 
     * UI widgets updated includes: Start/Stop Practice Session, TC toggle, 
     * Stop, and Volume control labels.
     * Also checks the current file, and if different updates the UI text and waveform.
     * This aligns with the MPQAudioService design and state diagram.
     */
	private void updateWidgetsLook() {
		Log.d(LOGTAG, "updateWidgetsLook() 00  current mState: " + mAudioServiceState + " mAWZoom: " + mAudioWaveformZoom);
		
		// Exit without updating UI, if the fragment is not attached to an activity.
		Activity la = getActivity();
		if (null == la) {
			Log.e(LOGTAG, "updateWidgetsLook() 05->99  EXIT , called with no current activity la:" + la + " mAct: " + mAct );
			return;
		}

		// Show, hide, and update text for various UI buttons.
		if (MPQAudioService.FLG_SRV_STATE_INIT == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(false);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.INVISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState)  {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.VISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mPracticeButton.setVisibility(View.VISIBLE);
			mPracticeButton.setText("Delete Audio");

//			mPracticeButton.setEnabled(true);

//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) {
			// Recording
			mTCButton.setText(R.string.button_aui_pause);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.INVISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_mic_control));
//			mPracticeButton.setText(getString(R.string.button_practice_inprocess));			
//			mPracticeButton.setEnabled(true);
		} else if (MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_rec);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.INVISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_mic_control));
//			mPracticeButton.setText(getString(R.string.button_practice_inprocess)); 
//			mPracticeButton.setEnabled(true); 
		} else if (MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_play);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.VISIBLE);

			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} else if (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) {
			mTCButton.setText(R.string.button_aui_pause);
			mTCButton.setEnabled(true);
			mStopButton.setEnabled(true);
			mLabelAudioPos.setVisibility(View.VISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} else if (MPQAudioService.FLG_SRV_STATE_STOP == mAudioServiceState) {
			// Leave previous mTCButton text.
			mTCButton.setEnabled(false);
			mStopButton.setEnabled(false);
			mLabelAudioPos.setVisibility(View.INVISIBLE);
			
			mLabelVolControl.setVisibility(View.GONE);
			mPracticeButton.setVisibility(View.GONE);
//			mLabelVolControl.setText(getString(R.string.label_aui_vol_control));
//			mPracticeButton.setText(getString(R.string.button_practice_ready)); 
//			mPracticeButton.setEnabled(false); 
		} 
		  else {
			Log.e(LOGTAG, "updateWidgetsLook() 80  (ignoring) UNHANDLED State: " + mAudioServiceState);
		}
		
		// For all states check the current file, and if different, update the UI text and waveform.
    	String lFileName = "";
    	if ( (null != mAudioService) &&  (true == mAudioServiceBound) ) {
        	lFileName = mAudioService.getFilenameBnd();
    		Log.d(LOGTAG, "updateWidgetsLook() 82  current filename: " + mAudioServiceFileName + " new filename: " + lFileName);
        	if ( (lFileName != null) && (lFileName != mAudioServiceFileName) ) {
        		mAudioServiceFileName = lFileName;
        		
        		// Update mAudioWaveformView widget. 
                mAudioWaveformMaxPos = 0;
                if (MPQAudioService.FLG_SRV_STATE_REC != mAudioServiceState) {
                	waveformLoadFromFile();
                }
                
                Log.w(LOGTAG, "updateWidgetsLook() 84  SoundFile: " + mAudioWaveformSoundFile + " mAWZoom: " + mAudioWaveformZoom);
                if (mAudioWaveformSoundFile != null) {
                    DisplayMetrics metrics = new DisplayMetrics();
                    mAct.getWindowManager().getDefaultDisplay().getMetrics(metrics);
                    mAudioWaveformDensity = metrics.density;
                    
                    mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
                    mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);
                    mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
                    if (mAudioWaveformZoom > 0) {
                    	mAudioWaveformView.setZoomLevel(mAudioWaveformZoom);
                    }
                }
        	} // END if ( (lFileName != null) && (lFileName != mAudioServiceFileName) )
        	        	
        	// Update file length and cursor positions (this is called infrequently).
        	long lFileLenPosMS = mAudioService.getFileLengthMSBnd();
    		Log.d(LOGTAG, "updateWidgetsLook() 86  current len: " + mAudioServiceMSLen + " new fileLen: " + lFileLenPosMS);
        	if ( (lFileLenPosMS >= 0) && (lFileLenPosMS != mAudioServiceMSLen) ) {
        		mAudioServiceMSLen = lFileLenPosMS;
        	}        	
        	lFileLenPosMS = mAudioService.getFilePosMSBnd();
        	Log.d(LOGTAG, "updateWidgetsLook() 87  current pos: " + mAudioServiceMSPos + " new pos: " + lFileLenPosMS);
        	if ( (lFileLenPosMS >= 0) && (lFileLenPosMS != mAudioServiceMSPos) ) {
        		mAudioServiceMSPos = lFileLenPosMS;
        	}        	
    	} // END if ( (null != mAudioService) &&  (true == mAudioServiceBound) )
    	
    	// Follow up with update of Audio related UI elements.
		updateAudioWidgetsLook();
		
	} // END updateWidgetsLook()
	
	
    /**
     * Update frequently changing Audio UI widgets, such as cursor position and volume.
     * {@link #mAudioServiceState} is checked for the current state.
     * {@link #mAudioServiceMSPos} is checked for the current position.
     * {@link #mAudioServiceVol} is checked for the current volume.
     * 
     * UI widgets updated includes: cursor position, file length, volume level.
     * This aligns with the MPQAudioService design and state diagram.
     */
	private void updateAudioWidgetsLook() {
		Log.d(LOGTAG, "updateAudioWidgetsLook() 00  current mState: " + mAudioServiceState + " pos: " + mAudioServiceMSPos + " len: " + mAudioServiceMSLen + " vol: " + mAudioServiceVol);

		if (MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) {
			mLabelAudioPos.setText( SessionsDB.longSessionMSecString(mAudioServiceMSPos) + "/" );
		} else if ( (MPQAudioService.FLG_SRV_STATE_REC == mAudioServiceState) || 
				(MPQAudioService.FLG_SRV_STATE_PAUSE_REC == mAudioServiceState) ) {
			// No mLabelAudioPos updates while recording; only audio file length.
		} else if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) ||
				(MPQAudioService.FLG_SRV_STATE_PAUSE_PLAY == mAudioServiceState) ) {
			mLabelAudioPos.setText( SessionsDB.longSessionMSecString(mAudioServiceMSPos) + "/");
		}
		
		mLabelAudioLen.setText( SessionsDB.longSessionMSecString(mAudioServiceMSLen) );
		//mLabelVolValue.setText(String.valueOf(mVolControl.getProgress()));
	} // END updateAudioWidgetsLook()	
	
	
	
	/*****************************************
	 * Waveform View Listener, followed by helper methods to update UI and load files. -- credit: Ringdroid
	 */
		
	/**
	 * Waveform UI widget listener.
	 */
	private WaveformView.WaveformListener waveformListener = new WaveformView.WaveformListener() {	
		/**
		 * Every time we get a message that our waveform drew, see if we need to
		 * animate and trigger another redraw.
		 */
		public void waveformDraw() {
	 //   	Log.d(LOGTAG, "waveformListener waveformDraw() 00 = serviceState: " + mAudioServiceState);
			mAudioWaveformWidth = mAudioWaveformView.getMeasuredWidth();
			if (mAudioWaveformOffsetGoal != mAudioWaveformOffset)
				waveformUpdateDisplay();
			else if (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) {
				waveformUpdateDisplay();
			}
			else if (mAudioWaveformFlingVelocity != 0) {
				waveformUpdateDisplay();
			}
		} // END waveformDraw()
	
		public void waveformZoomIn() {
			mAudioWaveformView.zoomIn();
			mAudioWaveformStartPos = mAudioWaveformView.getStart();
			mAudioWaveformEndPos = mAudioWaveformView.getEnd();
			mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
			mAudioWaveformOffset = mAudioWaveformView.getOffset();
			mAudioWaveformOffsetGoal = mAudioWaveformOffset;
			mAudioWaveformZoom = mAudioWaveformView.getZoomLevel();
			waveformUpdateDisplay();
		}
	
		public void waveformZoomOut() {
			mAudioWaveformView.zoomOut();
			mAudioWaveformStartPos = mAudioWaveformView.getStart();
			mAudioWaveformEndPos = mAudioWaveformView.getEnd();
			mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
			mAudioWaveformOffset = mAudioWaveformView.getOffset();
			mAudioWaveformOffsetGoal = mAudioWaveformOffset;
			mAudioWaveformZoom = mAudioWaveformView.getZoomLevel();
			waveformUpdateDisplay();
		}
	
		public void waveformTouchStart(float x) {
	    	Log.v(LOGTAG, "waveformListener waveformTouchStart() 00 = x: " + x);
			mAudioWaveformTouchDragging = true;
			mAudioWaveformTouchStart = x;
			mAudioWaveformTouchInitialOffset = mAudioWaveformOffset;
			mAudioWaveformFlingVelocity = 0;
			mAudioWaveformTouchStartMsec = System.currentTimeMillis();
		} // END waveformTouchStart()
	
		public void waveformTouchMove(float x) {
	    	//xxLog.d(LOGTAG, "waveformListener waveformTouchMove() 00 = x: " + x + " old mOffset: " + mAudioWaveformOffset);
	    	mAudioWaveformOffset = waveformBoundsCheck((int)(mAudioWaveformTouchInitialOffset + (mAudioWaveformTouchStart - x)));
			//xxLog.d(LOGTAG, "waveformListener waveformTouchMove() 07 new mOffset: " + mAudioWaveformOffset);
			waveformUpdateDisplay();
		}
	
		public void waveformTouchEnd() {
			Log.i(LOGTAG, "waveformListener waveformTouchEnd() 00 zoomlevel: " + mAudioWaveformZoom);
			mAudioWaveformTouchDragging = false;
			mAudioWaveformOffsetGoal = mAudioWaveformOffset;

			long elapsedMsec = System.currentTimeMillis() -	mAudioWaveformTouchStartMsec;
			if (elapsedMsec < 300) {	// Short tap.
				// FIXME: how/when pause; maybe with double-tap???
				
				// Check if playing; seek to new touched position.
				if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) ||
						(MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) ) { 
					Log.w(LOGTAG, "waveformListener waveformTouchEnd() 62 move cursor to touch: " + mAudioWaveformTouchStart + " offset: " + mAudioWaveformOffset);
					Intent intent = new Intent(mAct, MPQAudioService.class);
					intent.setAction(MPQAudioService.ACTION_SEEK);
					
					// Pass file position via "extra-bundle"  
					int seekMsec = mAudioWaveformView.pixelsToMillisecs( (int)(mAudioWaveformTouchStart + mAudioWaveformOffset) );
					
					Log.w(LOGTAG, "waveformListener waveformTouchEnd() 64 seekMsec sent from UI: " + seekMsec);
		        	Bundle lextraArgs = new Bundle();
		        	lextraArgs.putInt(MPQAudioService.ARG_FILEPOS, seekMsec);
		        	intent.putExtras(lextraArgs);
					
					mAct.startService(intent);	
					
					mAudioServiceMSPos = seekMsec;
					waveformUpdateDisplay();
/////		            mAudioWaveformView.setPlayback(seekMsec);
				}
			} // END if (elapsedMsec < 300)
		} // END waveformTouchEnd()
	
		public void waveformFling(float vx) {
			mAudioWaveformTouchDragging = false;
			mAudioWaveformOffsetGoal = mAudioWaveformOffset;
			mAudioWaveformFlingVelocity = (int)(-vx);
			waveformUpdateDisplay();
		}
	}; // END WaveformView.WaveformListener waveformListener
		
	
	/**
	 * Handle touch events and other parameters to update the display.
	 */
    private void waveformUpdateDisplay() {
    	//xxLog.d(LOGTAG, "waveformUpdateDisplay() 00 = serviceState: " + mAudioServiceState);

		if (!mAudioWaveformView.hasSoundFile()) {
			Log.e(LOGTAG, "waveformUpdateDisplay() 05->99  EXIT , no associated soundfile" );
			return;
		}
		
		if ( (MPQAudioService.FLG_SRV_STATE_PLAY == mAudioServiceState) || 
				(MPQAudioService.FLG_SRV_STATE_READY == mAudioServiceState) ) {
            int frames = mAudioWaveformView.millisecsToPixels( (int) mAudioServiceMSPos); // FIXME: hack cast over here -- clean this up
            mAudioWaveformView.setPlayback(frames);
            waveformSetOffsetGoalNoUpdate(frames - mAudioWaveformWidth / 2);
        }

        if (!mAudioWaveformTouchDragging) {
        	//xxLog.d(LOGTAG, "waveformUpdateDisplay() 11 !mAudioWaveformTouchDragging");
            int offsetDelta;

            if (mAudioWaveformFlingVelocity != 0) {
                offsetDelta = (int)(mAudioWaveformFlingVelocity / 30);
                if (mAudioWaveformFlingVelocity > 80) {
                    mAudioWaveformFlingVelocity -= 80;
                } else if (mAudioWaveformFlingVelocity < -80) {
                    mAudioWaveformFlingVelocity += 80;
                } else {
                    mAudioWaveformFlingVelocity = 0;
                }

                mAudioWaveformOffset += offsetDelta;

                if (mAudioWaveformOffset + mAudioWaveformWidth / 2 > mAudioWaveformMaxPos) {
                	mAudioWaveformOffset = mAudioWaveformMaxPos - mAudioWaveformWidth / 2;
                    mAudioWaveformFlingVelocity = 0;
                }
                if (mAudioWaveformOffset < 0) {
                	mAudioWaveformOffset = 0;
                    mAudioWaveformFlingVelocity = 0;
                }
                mAudioWaveformOffsetGoal = mAudioWaveformOffset;
            } else {
                offsetDelta = mAudioWaveformOffsetGoal - mAudioWaveformOffset;

                if (offsetDelta > 10)
                    offsetDelta = offsetDelta / 10;
                else if (offsetDelta > 0)
                    offsetDelta = 1;
                else if (offsetDelta < -10)
                    offsetDelta = offsetDelta / 10;
                else if (offsetDelta < 0)
                    offsetDelta = -1;
                else
                    offsetDelta = 0;

                mAudioWaveformOffset += offsetDelta;
            }
        } // END if (!mAudioWaveformTouchDragging) 

    	mAudioWaveformOffset = waveformBoundsCheck(mAudioWaveformOffset);
    	//xxLog.w(LOGTAG, "waveformUpdateDisplay() 50 = StartPos: " + mAudioWaveformStartPos + " EndPos: " + mAudioWaveformEndPos + " Offset: " + mAudioWaveformOffset);
        mAudioWaveformView.setParameters(mAudioWaveformStartPos, mAudioWaveformEndPos, mAudioWaveformOffset);
        mAudioWaveformView.invalidate();
    } // END waveformUpdateDisplay()

	  

    private void waveformLoadFromFile() {
    	 Log.d(LOGTAG, "waveformLoadFromFile() 00  ");

     	 mAudioWaveformLoadingLastUpdateTime = System.currentTimeMillis();
         mAudioWaveformLoadingKeepGoing = true;
         mAudioWaveformProgressDialog = new ProgressDialog(mAct );
         mAudioWaveformProgressDialog.setProgressStyle(ProgressDialog.STYLE_HORIZONTAL);
         mAudioWaveformProgressDialog.setTitle(R.string.progress_dialog_loading);
         mAudioWaveformProgressDialog.setCancelable(true);
         mAudioWaveformProgressDialog.setOnCancelListener(
             new DialogInterface.OnCancelListener() {
                 public void onCancel(DialogInterface dialog) {
                 	mAudioWaveformLoadingKeepGoing = false;
                 }
             });
         mAudioWaveformProgressDialog.show();

         final CheapSoundFile.ProgressListener listener =
             new CheapSoundFile.ProgressListener() {
                 public boolean reportProgress(double fractionComplete) {
                     //Log.d(LOGTAG, "CheapSoundFile.ProgressListener reportProgress() 00 fraction: " + fractionComplete + " keep going: " + mAudioWaveformLoadingKeepGoing);
                     long now = System.currentTimeMillis();
                     if (now - mAudioWaveformLoadingLastUpdateTime > 100) {
                         mAudioWaveformProgressDialog.setProgress(
                             (int)(mAudioWaveformProgressDialog.getMax() *
                                   fractionComplete));
                         mAudioWaveformLoadingLastUpdateTime = now;
                     }
                     return mAudioWaveformLoadingKeepGoing;
                 }
             };

         // Load the sound file in a background thread; a second child thread, triggered by the UI Handler, runs when file loading is finished.
         new Thread() { 
             public void run() { 
                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 00 to load cheap soundfile");
                 try {
                	 String lFileName = Environment.getExternalStorageDirectory().getAbsolutePath() + mAudioServiceFileName;
                	 mAudioWaveformSoundFile = CheapSoundFile.create(lFileName, listener, mAct);
                 } catch (final Exception e) {
                     Log.e(LOGTAG, "waveformLoadFromFile() thread run() Exception: " + e.getLocalizedMessage() );
                     e.printStackTrace();
                     // FIXME: handle this exception -- do we throw the exception again??? or do we kill/finish ourselves???
                 }
                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 60 got sound file, about to dismiss and call finish()");
                 mAudioWaveformProgressDialog.dismiss(); 
                 if (mAudioWaveformLoadingKeepGoing) {
                     Runnable runnable = new Runnable() {
                             public void run() {
                                 Log.d(LOGTAG, "waveformLoadFromFile() thread run() 88 SUB thread to call finishOpen()");
                                 waveformFinishLoadFromFile();
                             }
                         };
                     mFragHandler.post(runnable);
                 } 
             } // END run()
         }.start();
 		Log.d(LOGTAG, "waveformLoadFromFile() 99  ");
     } // END waveformLoadFromFile()

     private void waveformFinishLoadFromFile() {
  		Log.d(LOGTAG, "waveformFinishLoadFromFile() 00  file: " + mAudioServiceFileName);
  		
  		// Exit early if there is not a valid sound file.
  		if (null == mAudioWaveformSoundFile) {
			Log.e(LOGTAG, "waveformFinishLoadFromFile() 05->99  EXIT , NULL soundfile");
  			return;
  		}
  		
        mAudioWaveformView.setSoundFile(mAudioWaveformSoundFile);
        mAudioWaveformView.recomputeHeights(mAudioWaveformDensity);

        mAudioWaveformMaxPos = mAudioWaveformView.maxPos();
        
        if (mAudioWaveformZoom > 0) {
        	mAudioWaveformView.setZoomLevel(mAudioWaveformZoom);
        }

        mAudioWaveformTouchDragging = false;
        mAudioWaveformOffset = 0;
        mAudioWaveformOffsetGoal = 0;
        mAudioWaveformFlingVelocity = 0;
        waveformResetPositions();
        if (mAudioWaveformEndPos > mAudioWaveformMaxPos) {
         	mAudioWaveformEndPos = mAudioWaveformMaxPos;
        }

        updateWidgetsLook();
   		Log.d(LOGTAG, "waveformFinishLoadFromFile() 99");
     } // END waveformFinishLoadFromFile()
 	
     private void waveformResetPositions() {
     	mAudioWaveformStartPos = mAudioWaveformView.secondsToPixels(0.0);
     	mAudioWaveformEndPos = mAudioWaveformView.getEnd();
     }
    
     /**
      * Check bounds on pos, ensuring it is within the available audio.
      * @param pos int with position in audio.  Unit is int value of frames, corresponding with current zoom level.
      * @return pos int value which has been validated.
      */
     private int waveformBoundsCheck(int pos) {
         if (pos < 0)
             return 0;
         if (pos > mAudioWaveformMaxPos)
             return mAudioWaveformMaxPos;
         return pos;
     }
     
     private void waveformSetOffsetGoalNoUpdate(int offset) {
         if (mAudioWaveformTouchDragging) {
             return;
         }

         mAudioWaveformOffsetGoal = offset;
         if (mAudioWaveformOffsetGoal + mAudioWaveformWidth / 2 > mAudioWaveformMaxPos)
        	 mAudioWaveformOffsetGoal = mAudioWaveformMaxPos - mAudioWaveformWidth / 2;
         if (mAudioWaveformOffsetGoal < 0)
        	 mAudioWaveformOffsetGoal = 0;
     }
     
} // END class MPQAudioFragment
